@@ -23,12 +23,37 @@ PROGRAM = 'ubuntu-image'
 
 class SizeAction(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
-        try:
-            size = as_size(values)
-        except (KeyError, ValueError):
-            raise argparse.ArgumentError(
-                self, 'Invalid size: {}'.format(values))
-        setattr(namespace, self.dest, size)
+        sizes = {}
+        specs = values.split(',')
+        # First check for extended syntax by splitting on commas.
+        for spec in specs:
+            index, colon, size = spec.partition(':')
+            if colon != ':':
+                if len(specs) != 1:
+                    raise argparse.ArgumentError(
+                        self,
+                        'Invalid multi-volume size specification: {}'.format(
+                            spec))
+                # Backward compatibility.  Since there was no colon to
+                # partition the string on, the size argument will be in the
+                # index local variable.
+                try:
+                    sizes = as_size(index)
+                except (KeyError, ValueError):
+                    raise argparse.ArgumentError(
+                        self, 'Invalid size: {}'.format(values))
+                break
+            try:
+                size = as_size(size)
+            except (KeyError, ValueError):
+                raise argparse.ArgumentError(
+                    self, 'Invalid size: {}'.format(values))
+            try:
+                index = int(index)
+            except (ValueError, TypeError):
+                pass
+            sizes[index] = size
+        setattr(namespace, self.dest, sizes)
         # For display purposes.
         namespace.given_image_size = values
 
@@ -53,19 +78,31 @@ def parseargs(argv=None):
         default=False, action='store_true',
         help=_('Enable debugging output'))
     common_group.add_argument(
+        '-i', '--image-size',
+        default=None, action=SizeAction, metavar='SIZE',
+        help=_("""The suggested size of the generated disk image file.  If this
+        size is smaller than the minimum calculated size of the image a warning
+        will be issued and --image-size will be ignored.  The value is the size
+        in bytes, with allowable suffixes 'M' for MiB and 'G' for GiB.  Use an
+        extended syntax to define the suggested size for the disk images
+        generated by a multi-volume gadget.yaml spec.  See the ubuntu-image(1)
+        manpage for details."""))
+    output_group = common_group.add_mutually_exclusive_group()
+    output_group.add_argument(
+        '-O', '--output-dir',
+        default=None, metavar='DIRECTORY',
+        help=_("""The directory in which to put generated disk image files.
+        The disk image files themselves will be named <volume>.img inside this
+        directory, where <volume> is the volume name taken from the
+        gadget.yaml file.  Use this option instead of the deprecated
+        -o/--output option."""))
+    output_group.add_argument(
         '-o', '--output',
         default=None, metavar='FILENAME',
-        help=_("""The generated disk image file.  If not given, the image will
-        be put in a file called disk.img in the working directory (in which
-        case, you probably want to specify -w)."""))
-    common_group.add_argument(
-        '--image-size',
-        default=None, action=SizeAction, metavar='SIZE',
-        help=_("""The size of the generated disk image file (see
-        -o/--output).  If this size is smaller than the minimum calculated
-        size of the image a warning will be issued and --image-size will be
-        ignored.  The value is the size in bytes, with allowable suffixes 'M'
-        for MiB and 'G' for GiB."""))
+        help=_("""DEPRECATED (use -O/--output-dir instead).  The generated
+        disk image file.  If not given, the image will be put in a file called
+        disk.img in the working directory (in which case, you probably want to
+        specify -w)."""))
     # Snap-based image options.
     snap_group = parser.add_argument_group(
         _('Image contents options'),
@@ -133,6 +170,10 @@ def parseargs(argv=None):
         args.thru = int(args.thru)
     with suppress(ValueError, TypeError):
         args.until = int(args.until)
+    # -o/--output is deprecated and mutually exclusive with -O/--output-dir
+    if args.output is not None:
+        print('-o/--output is deprecated; use -O/--output-dir instead',
+              file=sys.stderr)
     return args
 
 
@@ -151,7 +192,7 @@ def main(argv=None):
         try:
             state_machine = ModelAssertionBuilder(args)
         except TMPNotReadableFromOutsideSnap as error:
-            print(error.__doc__, file=sys.stderr)
+            print(str(error), file=sys.stderr)
             return 1
     # Run the state machine, either to the end or thru/until the named state.
     try:
