@@ -5,12 +5,14 @@ import shutil
 import logging
 
 from math import ceil
+from pathlib import Path
 from subprocess import CalledProcessError
 from tempfile import TemporaryDirectory
-from ubuntu_image.helpers import MiB, mkfs_ext4, run, snap, sparse_copy
-from ubuntu_image.image import Image, MBRImage
+from ubuntu_image.helpers import MiB, mkfs_ext4, run, snap
+from ubuntu_image.image import Image
 from ubuntu_image.parser import (
-    BootLoader, FileSystemType, VolumeSchema, parse as parse_yaml)
+    BootLoader, FileSystemType, StructureRole, VolumeSchema,
+    parse as parse_yaml)
 from ubuntu_image.state import State
 
 
@@ -18,6 +20,10 @@ SPACE = ' '
 _logger = logging.getLogger('ubuntu-image')
 
 
+class TMPNotReadableFromOutsideSnap(Exception):
+    """/tmp is different from inside and outside a snap."""
+
+
 class ModelAssertionBuilder(State):
     def __init__(self, args):
         super().__init__()
@@ -34,25 +40,52 @@ class ModelAssertionBuilder(State):
             self.resources.enter_context(TemporaryDirectory())
             if args.workdir is None
             else args.workdir)
-        # Where the disk.img file ends up.
-        self.output = (
-            os.path.join(self.workdir, 'disk.img')
-            if args.output is None
-            else args.output)
+        # Where the disk.img file ends up.  /tmp to a snap is not the same
+        # /tmp outside of the snap.  When running as a snap, don't allow the
+        # user to output a disk image to a location that won't exist for them.
+        # When run as a snap, /tmp is not writable.
+        if any(key.startswith('SNAP') for key in os.environ):
+            # The output directories, in order of precedence.
+            check_paths = (args.output, args.output_dir, os.getcwd())
+            # This loop will never exit normally because either we'll hit a
+            # /tmp directory, or we won't.  In the former case we'll always
+            # exit by raising the exception, and in the latter case, we'll hit
+            # the break.  Therefore, tell coverage not to check partial
+            # branches for this loop.
+            for path in check_paths:                # pragma: no branch
+                if path is None:
+                    continue
+                path = os.sep.join(path.split(os.sep)[:2])
+                if path == '/tmp':
+                    raise TMPNotReadableFromOutsideSnap(
+                        'ubuntu-image snap cannot write images to /tmp')
+                # This path is okay and since it'll take precedence, we're
+                # done checking.
+                break
+            # Check the location of the model assertion.
+            path = os.sep.join(args.model_assertion.split(os.sep)[:2])
+            if path == '/tmp':
+                raise TMPNotReadableFromOutsideSnap(
+                    'ubuntu-image snap cannot read models from /tmp')
+            # Check all the extra snaps.
+            extra_snaps = [] if args.extra_snaps is None else args.extra_snaps
+            for snap_path in extra_snaps:
+                path = os.sep.join(snap_path.split(os.sep)[:2])
+                if path == '/tmp':
+                    raise TMPNotReadableFromOutsideSnap(
+                        'ubuntu-image snap cannot read extra snaps from /tmp')
+        self.output_dir = args.output_dir
+        self.output = args.output
         # Information passed between states.
         self.rootfs = None
         self.rootfs_size = 0
-        self.rootfs_offset = 0
-        self.image_size = 0
-        self.bootfs = None
-        self.bootfs_sizes = None
+        self.part_images = None
         self.images = None
-        self.boot_images = None
-        self.root_img = None
-        self.disk_img = None
+        self.entry = None
         self.gadget = None
         self.args = args
         self.unpackdir = None
+        self.volumedir = None
         self.cloud_init = args.cloud_init
         self.exitcode = 0
         self._next.append(self.make_temporary_directories)
@@ -61,42 +94,34 @@ class ModelAssertionBuilder(State):
         state = super().__getstate__()
         state.update(
             args=self.args,
-            boot_images=self.boot_images,
-            bootfs=self.bootfs,
-            bootfs_sizes=self.bootfs_sizes,
             cloud_init=self.cloud_init,
-            disk_img=self.disk_img,
             exitcode=self.exitcode,
             gadget=self.gadget,
-            image_size=self.image_size,
             images=self.images,
             output=self.output,
-            root_img=self.root_img,
+            output_dir=self.output_dir,
+            part_images=self.part_images,
             rootfs=self.rootfs,
             rootfs_size=self.rootfs_size,
-            rootfs_offset=self.rootfs_offset,
             unpackdir=self.unpackdir,
+            volumedir=self.volumedir,
             )
         return state
 
     def __setstate__(self, state):
         super().__setstate__(state)
         self.args = state['args']
-        self.boot_images = state['boot_images']
-        self.bootfs = state['bootfs']
-        self.bootfs_sizes = state['bootfs_sizes']
         self.cloud_init = state['cloud_init']
-        self.disk_img = state['disk_img']
         self.exitcode = state['exitcode']
         self.gadget = state['gadget']
-        self.image_size = state['image_size']
         self.images = state['images']
         self.output = state['output']
-        self.root_img = state['root_img']
+        self.output_dir = state['output_dir']
+        self.part_images = state['part_images']
         self.rootfs = state['rootfs']
         self.rootfs_size = state['rootfs_size']
-        self.rootfs_offset = state['rootfs_offset']
         self.unpackdir = state['unpackdir']
+        self.volumedir = state['volumedir']
 
     def _log_exception(self, name):
         # Only log the exception if we're in debug mode.
@@ -106,6 +131,7 @@ class ModelAssertionBuilder(State):
     def make_temporary_directories(self):
         self.rootfs = os.path.join(self.workdir, 'root')
         self.unpackdir = os.path.join(self.workdir, 'unpack')
+        self.volumedir = os.path.join(self.workdir, 'volumes')
         os.makedirs(self.rootfs)
         # Despite the documentation, `snap prepare-image` doesn't create the
         # gadget/ directory.
@@ -131,6 +157,20 @@ class ModelAssertionBuilder(State):
         shutil.copy(yaml_file, self.workdir)
         with open(yaml_file, 'r', encoding='utf-8') as fp:
             self.gadget = parse_yaml(fp)
+        # Make a working subdirectory for every volume we're going to create.
+        # We'll put the volume contents inside these directories, and then use
+        # the directories to create the disk images, one per volume.
+        #
+        # Store some additional metadata on the VolumeSpec object.  This is
+        # convenient, if crufty, since we're poking data onto an object from
+        # the outside.
+        for name, volume in self.gadget.volumes.items():
+            volume.basedir = os.path.join(self.volumedir, name)
+            os.makedirs(volume.basedir)
+        envar = os.environ.get('UBUNTU_IMAGE_PRESERVE_UNPACK')
+        if envar is not None:
+            preserve_dir = os.path.join(envar, 'unpack')
+            shutil.copytree(self.unpackdir, preserve_dir)
         self._next.append(self.populate_rootfs_contents)
 
     def populate_rootfs_contents(self):
@@ -177,29 +217,18 @@ class ModelAssertionBuilder(State):
         self._next.append(self.pre_populate_bootfs_contents)
 
     def pre_populate_bootfs_contents(self):
-        volumes = self.gadget.volumes.values()
-        assert len(volumes) == 1, 'For now, only one volume is allowed'
-        volume = list(volumes)[0]
-        for partnum, part in enumerate(volume.structures):
-            target_dir = os.path.join(self.workdir, 'part{}'.format(partnum))
-            os.makedirs(target_dir, exist_ok=True)
+        for name, volume in self.gadget.volumes.items():
+            for partnum, part in enumerate(volume.structures):
+                target_dir = os.path.join(
+                    volume.basedir, 'part{}'.format(partnum))
+                os.makedirs(target_dir, exist_ok=True)
         self._next.append(self.populate_bootfs_contents)
 
-    def populate_bootfs_contents(self):
-        # XXX We currently support only one volume specification.
-        assert len(self.gadget.volumes) == 1, (
-            'For now, only one volume is allowed')
-        # The unpack directory has a boot/ directory inside it.  The contents
-        # of this directory (but not the parent <unpack>/boot directory
-        # itself) needs to be moved to the bootfs directory.
-        volume = list(self.gadget.volumes.values())[0]
-        # At least one structure is required.
+    def _populate_one_bootfs(self, name, volume):
         for partnum, part in enumerate(volume.structures):
-            target_dir = os.path.join(self.workdir, 'part{}'.format(partnum))
-            # XXX: Use file system label for the moment, until we get a proper
-            # way to identify the boot partition.
-            if part.filesystem_label == 'system-boot':
-                self.bootfs = target_dir
+            target_dir = os.path.join(volume.basedir, 'part{}'.format(partnum))
+            if part.role is StructureRole.system_boot:
+                volume.bootfs = target_dir
                 if volume.bootloader is BootLoader.uboot:
                     boot = os.path.join(
                         self.unpackdir, 'image', 'boot', 'uboot')
@@ -259,84 +288,105 @@ class ModelAssertionBuilder(State):
                         # a useful error message instead of a traceback.
                         os.makedirs(os.path.dirname(dst), exist_ok=True)
                         shutil.copy(src, dst)
-        self._next.append(self.calculate_bootfs_size)
 
-    def calculate_bootfs_size(self):
-        volumes = self.gadget.volumes.values()
-        assert len(volumes) == 1, 'For now, only one volume is allowed'
-        volume = list(volumes)[0]
-        self.bootfs_sizes = {}
-        # At least one structure is required.
-        for i, part in enumerate(volume.structures):
-            if part.filesystem is FileSystemType.none:
-                continue
-            partnum = 'part{}'.format(i)
-            target_dir = os.path.join(self.workdir, partnum)
-            self.bootfs_sizes[partnum] = self._calculate_dirsize(target_dir)
+    def populate_bootfs_contents(self):
+        for name, volume in self.gadget.volumes.items():
+            self._populate_one_bootfs(name, volume)
         self._next.append(self.prepare_filesystems)
 
+    def _prepare_one_volume(self, i, name, volume):
+        volume.part_images = []
+        farthest_offset = 0
+        for partnum, part in enumerate(volume.structures):
+            part_img = os.path.join(
+                volume.basedir, 'part{}.img'.format(partnum))
+            if part.role is StructureRole.system_data:
+                # The image for the root partition.
+                if part.size is None:
+                    part.size = self.rootfs_size
+                elif part.size < self.rootfs_size:
+                    _logger.warning('rootfs partition size ({}) smaller than '
+                                    'actual rootfs contents {}'.format(
+                                        part.size, self.rootfs_size))
+                    part.size = self.rootfs_size
+                # We defer creating the root file system image because we have
+                # to populate it at the same time.  See mkfs.ext4(8) for
+                # details.
+                Path(part_img).touch()
+                os.truncate(part_img, self.rootfs_size)
+            else:
+                run('dd if=/dev/zero of={} count=0 bs={} seek=1'.format(
+                    part_img, part.size))
+                if part.filesystem is FileSystemType.vfat:
+                    label_option = (
+                        '-n {}'.format(part.filesystem_label)
+                        # XXX I think this could be None or the empty string,
+                        # but this needs verification.
+                        if part.filesystem_label
+                        else '')
+                    # XXX: hard-coding of sector size.
+                    run('mkfs.vfat -s 1 -S 512 -F 32 {} {}'.format(
+                        label_option, part_img))
+            volume.part_images.append(part_img)
+            farthest_offset = max(farthest_offset, (part.offset + part.size))
+        # Calculate or check the final image size.
+        #
+        # XXX: Hard-codes last 34 512-byte sectors for backup GPT,
+        # empirically derived from sgdisk behavior.
+        calculated = ceil(farthest_offset / 1024 + 17) * 1024
+        if self.args.image_size is None:
+            volume.image_size = calculated
+        elif isinstance(self.args.image_size, int):
+            # One size to rule them all.
+            if self.args.image_size < calculated:
+                _logger.warning(
+                    'Ignoring image size smaller '
+                    'than minimum required size: vol[{}]:{} '
+                    '{} < {}'.format(
+                        i, name, self.args.given_image_size, calculated))
+                volume.image_size = calculated
+            else:
+                volume.image_size = self.args.image_size
+        else:
+            # The --image-size arguments are a dictionary, so look up the
+            # one used for this volume.
+            size_by_index = self.args.image_size.get(i)
+            size_by_name = self.args.image_size.get(name)
+            if size_by_index is not None and size_by_name is not None:
+                _logger.warning(
+                    'Ignoring ambiguous volume size; index+name given')
+                volume.image_size = calculated
+            else:
+                image_size = (size_by_index
+                              if size_by_name is None
+                              else size_by_name)
+                if image_size < calculated:
+                    _logger.warning(
+                        'Ignoring image size smaller '
+                        'than minimum required size: vol[{}]:{} '
+                        '{} < {}'.format(
+                            i, name, self.args.given_image_size, calculated))
+                    volume.image_size = calculated
+                else:
+                    volume.image_size = image_size
+
     def prepare_filesystems(self):
         self.images = os.path.join(self.workdir, '.images')
         os.makedirs(self.images)
-        # The image for the boot partition.
-        self.boot_images = []
-        volumes = self.gadget.volumes.values()
-        assert len(volumes) == 1, 'For now, only one volume is allowed'
-        volume = list(volumes)[0]
-        for partnum, part in enumerate(volume.structures):
-            # The root file system implicitly lives right after the farthest
-            # out structure, which may not be the last named structure in the
-            # gadget.yaml.
-            self.rootfs_offset = max(
-                self.rootfs_offset, (part.offset + part.size))
-            part_img = os.path.join(self.images, 'part{}.img'.format(partnum))
-            self.boot_images.append(part_img)
-            run('dd if=/dev/zero of={} count=0 bs={} seek=1'.format(
-                part_img, part.size))
-            if part.filesystem is FileSystemType.vfat:
-                label_option = (
-                    '-n {}'.format(part.filesystem_label)
-                    # XXX I think this could be None or the empty string, but
-                    # this needs verification.
-                    if part.filesystem_label
-                    else '')
-                # XXX: hard-coding of sector size
-                run('mkfs.vfat -s 1 -S 512 -F 32 {} {}'.format(
-                    label_option, part_img))
-        # Sanity check whether everything will fit on the disk if an explicit
-        # size was given on the command line.  XXX: This hard-codes last 34
-        # 512-byte sectors for backup GPT, empirically derived from sgdisk
-        # behavior.
-        calculated = ceil(
-            (self.rootfs_offset + self.rootfs_size) / 1024 + 17) * 1024
-        if self.args.image_size is None:
-            self.image_size = calculated
-        else:
-            if self.args.image_size < calculated:
-                _logger.warning('Ignoring --image-size={} smaller '
-                                'than minimum required size {}'.format(
-                                    self.args.given_image_size, calculated))
-                self.image_size = calculated
-            else:
-                self.image_size = self.args.image_size
-        # The image for the root partition.
-        self.root_img = os.path.join(self.images, 'root.img')
-        # Create empty file with holes.
-        with open(self.root_img,  'w'):
-            pass
-        os.truncate(self.root_img, self.rootfs_size)
-        # We defer creating the root file system image because we have to
-        # populate it at the same time.  See mkfs.ext4(8) for details.
+        for i, (name, volume) in enumerate(self.gadget.volumes.items()):
+            self._prepare_one_volume(i, name, volume)
         self._next.append(self.populate_filesystems)
 
-    def populate_filesystems(self):
-        volumes = self.gadget.volumes.values()
-        assert len(volumes) == 1, 'For now, only one volume is allowed'
-        volume = list(volumes)[0]
+    def _populate_one_volume(self, name, volume):
         for partnum, part in enumerate(volume.structures):
-            part_img = self.boot_images[partnum]
-            part_dir = os.path.join(self.workdir, 'part{}'.format(partnum))
-            if part.filesystem is FileSystemType.none:
+            part_img = volume.part_images[partnum]
+            part_dir = os.path.join(volume.basedir, 'part{}'.format(partnum))
+            if part.role is StructureRole.system_data:
+                # The root partition needs to be ext4, which may or may not be
+                # populated at creation time, depending on the version of
+                # e2fsprogs.
+                mkfs_ext4(part_img, self.rootfs, part.filesystem_label)
+            elif part.filesystem is FileSystemType.none:
                 image = Image(part_img, part.size)
                 offset = 0
                 for content in part.content:
@@ -367,14 +417,83 @@ class ModelAssertionBuilder(State):
             else:
                 raise AssertionError('Invalid part filesystem type: {}'.format(
                     part.filesystem))
-        # The root partition needs to be ext4, which may or may not be
-        # populated at creation time, depending on the version of e2fsprogs.
-        mkfs_ext4(self.root_img, self.rootfs)
+
+    def populate_filesystems(self):
+        for name, volume in self.gadget.volumes.items():
+            self._populate_one_volume(name, volume)
         self._next.append(self.make_disk)
 
-    def make_disk(self):
-        self.disk_img = os.path.join(self.images, 'disk.img')
+    def _make_one_disk(self, imgfile, name, volume):
         part_id = 1
+        # Create the image object for the selected volume schema
+        image = Image(imgfile, volume.image_size, volume.schema)
+        offset_writes = []
+        part_offsets = {}
+        # We first create all the partitions.
+        for part in volume.structures:
+            if part.name is not None:
+                part_offsets[part.name] = part.offset
+            if part.offset_write is not None:
+                offset_writes.append((part.offset, part.offset_write))
+            if part.role is StructureRole.mbr or part.type == 'bare':
+                continue
+            activate = False
+            if (volume.schema is VolumeSchema.mbr and
+                    part.role is StructureRole.system_boot):
+                activate = True
+            elif (volume.schema is VolumeSchema.gpt and
+                    part.role is StructureRole.system_data and
+                    part.name is None):
+                part.name = 'writable'
+            image.partition(part.offset, part.size, part.name, activate)
+        # Now since we're done, we need to do a second pass to copy the data
+        # and set all the partition types.  This needs to be done like this as
+        # libparted's commit() operation resets type GUIDs to defaults and
+        # clobbers things like hybrid MBR partitions.
+        part_id = 1
+        for i, part in enumerate(volume.structures):
+            image.copy_blob(volume.part_images[i],
+                            bs=image.sector_size,
+                            seek=part.offset // image.sector_size,
+                            count=ceil(part.size / image.sector_size),
+                            conv='notrunc')
+            if part.role is StructureRole.mbr or part.type == 'bare':
+                continue
+            image.set_parition_type(part_id, part.type)
+            part_id += 1
+        for value, dest in offset_writes:
+            # Decipher non-numeric offset_write values.
+            if isinstance(dest, tuple):
+                dest = part_offsets[dest[0]] + dest[1]
+            image.write_value_at_offset(value // image.sector_size, dest)
+
+    def make_disk(self):
+        # Based on the -o/--output and -O/--output-dir options, and the volumes
+        # in the gadget.yaml file, we can now calculate where the generated
+        # disk images should go.  We'll write them directly to the final
+        # destination so they don't have to be moved later.  Here is the
+        # option precedence:
+        #
+        # * The location specified by -o/--output;
+        # * <output_dir>/<volume_name>.img
+        # * <work_dir>/disk.img
+        #
+        # If -o was given and there are multiple volumes, we ignore it and
+        # act as if -O is in use.
+        disk_img = output_dir = None
+        if self.output is not None:
+            if len(self.gadget.volumes) > 1:
+                _logger.warn('-o/--output ignored for multiple volumes')
+            else:
+                disk_img = self.output
+        # The argument parser ensures that these are mutually exclusive.
+        if disk_img is None:
+            if self.output_dir is None:
+                output_dir = (os.getcwd() if self.args.workdir is None
+                              else self.args.workdir)
+            else:
+                output_dir = self.output_dir
+            os.makedirs(output_dir, exist_ok=True)
         # Walk through all partitions and write them to the disk image at the
         # lowest permissible offset.  We should not have any overlapping
         # partitions, the parser should have already rejected such as invalid.
@@ -382,74 +501,12 @@ class ModelAssertionBuilder(State):
         # XXX: The parser should sort these partitions for us in disk order as
         # part of checking for overlaps, so we should not need to sort them
         # here.
-        volumes = self.gadget.volumes.values()
-        assert len(volumes) == 1, 'For now, only one volume is allowed'
-        volume = list(volumes)[0]
-        # XXX: This ought to be a single constructor that figures out the
-        # class for us when we pass in the schema.
-        if volume.schema is VolumeSchema.mbr:
-            image = MBRImage(self.disk_img, self.image_size)
-        else:
-            image = Image(self.disk_img, self.image_size)
-        offset_writes = []
-        part_offsets = {}
-        root_offset = 0
-        for i, part in enumerate(volume.structures):
-            if part.name is not None:
-                part_offsets[part.name] = part.offset
-            if part.offset_write is not None:
-                offset_writes.append((part.offset, part.offset_write))
-            image.copy_blob(self.boot_images[i],
-                            bs='1M', seek=part.offset // MiB(1),
-                            count=ceil(part.size / MiB(1)),
-                            conv='notrunc')
-            if part.type in ('bare', 'mbr'):
-                continue
-            # sgdisk takes either a sector or a KiB/MiB argument; assume
-            # that the offset and size are always multiples of 1MiB.
-            #
-            # XXX Size must not be zero, which will happen if part.size < 1MiB
-            partition_args = dict(
-                new='{}M:+{}M'.format(
-                    part.offset // MiB(1), part.size // MiB(1)),
-                typecode=part.type,
-                )
-            # XXX: special-casing.
-            if (volume.schema is VolumeSchema.mbr and
-                    part.filesystem_label == 'system-boot'):
-                partition_args['activate'] = True
-            if part.name is not None:
-                partition_args['change_name'] = part.name
-            image.partition(part_id, **partition_args)
-            part_id += 1
-            # Put the rootfs after the farthest out structure, which may not
-            # be the last named structure in the gadget.yaml.
-            root_offset = max(root_offset, (part.offset + part.size))
-        # Create and write the main snappy writable partition (i.e. rootfs) as
-        # the last partition.
-        root_offset_mib = self.rootfs_offset // MiB(1)
-        root_size_kib = ceil(self.rootfs_size / 1024)
-        image.partition(
-            part_id,
-            new='{}M:+{}K'.format(root_offset_mib, root_size_kib),
-            typecode=('83', '0FC63DAF-8483-4772-8E79-3D69D8477DE4'))
-        if volume.schema is VolumeSchema.gpt:
-            image.partition(part_id, change_name='writable')
-        image.copy_blob(
-            self.root_img,
-            bs='1M', seek=root_offset_mib,
-            count=ceil(self.rootfs_size / MiB(1)),
-            conv='notrunc')
-        for value, dest in offset_writes:
-            # Decipher non-numeric offset_write values.
-            if isinstance(dest, tuple):
-                dest = part_offsets[dest[0]] + dest[1]
-            # XXX: Hard-coding of 512-byte sectors.
-            image.write_value_at_offset(value // 512, dest)
+        for name, volume in self.gadget.volumes.items():
+            image_path = (
+                disk_img if disk_img is not None
+                else os.path.join(output_dir, '{}.img'.format(name)))
+            self._make_one_disk(image_path, name, volume)
         self._next.append(self.finish)
 
     def finish(self):
-        # Move the completed disk image to destination location, since the
-        # temporary scratch directory is about to get removed.
-        shutil.move(self.disk_img, self.output, copy_function=sparse_copy)
         self._next.append(self.close)
